require "metasm"
require "pp"

class SSTIC14_MCU_CPU < Metasm::CPU
	class Reg
		include Metasm::Renderable

		def initialize(i);  @i = i; end
		def symbolic; "r#@i".to_sym; end
		def render; ["r#@i"]; end
	end

	class Cond
		include Metasm::Renderable
        attr_reader :i
        Flags = ["z", "1", "2", "3", "4", "5", "6", "lte", "lt", "9", "a", "b", "gt" , "d", "e", "nz"]

		def initialize(i);  @i = i; end
		def symbolic; Flags[@i].to_sym; end
		def render; [Flags[@i]]; end
	end

	class MemRef
		include Metasm::Renderable
		def initialize(ptr,disp=nil); @ptr = ptr; @disp=disp end
		def symbolic;
            Metasm::Indirection[@ptr.kind_of?(Metasm::Expression) ? @ptr : @ptr.symbolic, 1]; 
        end
		def render; 
            res = ['[', @ptr, '+', @disp, ']' ]
        end
	end

	def initialize
		@size = 32
		@endianness = :little
	end

	def addop(name, bin, *args)
		o = Metasm::Opcode.new(name, bin)
		args.each { |a|
			o.args << a if @valid_args[a]
			o.props[a] = true if @valid_props[a]
			o.fields[a] = [@fields_mask[a], @fields_shift[a]] if @fields_mask[a]
		}
		@opcode_list << o
	end

	def init_opcode_list
		@opcode_list = []

		@valid_props = { :setip => true, :stopexec => true, :byteop => true}
		@fields_mask = { :r_d => 0xF, :r_s => 0xF, :r_s2 => 0xF, :i8 => 0xFF, :i16 => 0xFFFF, :flags => 0xF, :disp => 0xFF0F, :disp8 => 0xFF}
		@fields_shift = { :r_d => 0, :r_s => 8, :r_s2 => 12, :i8 => 8, :flags => 0, :i16 => 8, :disp => 0, :disp8 => 8}
		@valid_args = { :r_d => true, :r_s => true, :r_s2 => true, :i8 => true, :i16 => true, :flags => true, :disp=> true, :disp8 => true}

        # 4 bytes opcodes
		addop 'll',  0x10, :r_d, :i8
		addop 'lh',  0x20, :r_d, :i8
        # macro instruction
		addop 'li', 0, :r_d, :i16
		addop 'xor',  0x30, :r_d, :r_s, :r_s2
		addop 'add',  0x40, :r_d, :r_s, :r_s2
		addop 'mov',  0x50, :r_d, :r_s, :r_s2
		addop 'add',  0x60, :r_d, :r_s, :r_s2
		addop 'sub',  0x70, :r_d, :r_s, :r_s2
		addop 'mul',  0x80, :r_d, :r_s, :r_s2
		addop 'mod',  0x90, :r_d, :r_s, :r_s2
		addop 'j', 0xa0, :flags, :disp8, :setip
		addop 'jmp', 0xb0, :disp8, :setip, :stopexec
#		addop 'jmp', 0xb3, :byteop, :disp8, :setip, :stopexec
#		addop 'jmp', 0xB0, :disp8,:r_s,:setip
		addop 'syscall', 0xc8, :byteop, :i8
		addop 'call', 0xc0, :disp, :setip
		addop 'call', 0xd0, :r_s, :i8
		addop 'ret', 0xd0, :stopexec
		addop 'movf',  0xE0, :r_d, :r_s, :r_s2
		addop 'movt',  0xF0, :r_d, :r_s, :r_s2

	end
	
	def build_bin_lookaside
		opcode_list.each { |op|
			mask = 0
			op.fields.each { |fld, (msk, shft)| mask |= msk << shft }
            if not op.props[:byteop]
                mask |= 0xF
            end
			op.bin_mask = mask^0xFFFF
           #puts ("%s %04x %04x" % [op.name, op.bin, op.bin_mask])
		}
	end

	def decode_findopcode(edata)
		di = Metasm::DecodedInstruction.new(self)
		val = edata.decode_imm(:u16, @endianness)
		edata.ptr -= 2
		if op = opcode_list.find { |o| val & o.bin_mask == o.bin }
			di.opcode = op
            di.bin_length = 2
        else
            puts "%04x" % val
            raise "fuu"
		end
        di
	end

	def decode_instr_op(edata, di)
        val = edata.decode_imm(:u16, @endianness)
		di.instruction.opname = di.opcode.name

        case di.opcode.name 
            when "movf" then begin
                di.instruction.args << Reg.new((val >> @fields_shift[:r_d]) & @fields_mask[:r_d])
                di.instruction.args << MemRef.new(Reg.new((val >> @fields_shift[:r_s]) & @fields_mask[:r_s]), Reg.new((val >> @fields_shift[:r_s2]) & @fields_mask[:r_s2]))
            end
            when "movt" then begin
                di.instruction.args << MemRef.new(Reg.new((val >> @fields_shift[:r_s]) & @fields_mask[:r_s]), Reg.new((val >> @fields_shift[:r_s2]) & @fields_mask[:r_s2]))
                di.instruction.args << Reg.new((val >> @fields_shift[:r_d]) & @fields_mask[:r_d])
            end
            when "jmp" then begin
                # negative disp8
                if val&0xFF == 0xB3
                    val = (val >> @fields_shift[:disp8]) & @fields_mask[:disp8]
                    val = -((val^0xFF)+1)
                else
                    val = (val >> @fields_shift[:disp8]) & @fields_mask[:disp8]
                end
                di.instruction.args << Metasm::Expression[val+2]
            end
            else begin
                di.opcode.args.each { |a|
                    di.instruction.args << case a
                    when :r_s2; Reg.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :r_s; Reg.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :r_d; Reg.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :flags; Cond.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :i8; Metasm::Expression[(val >> @fields_shift[a]) & @fields_mask[a]]
                    when :disp then begin
                        val = (val >> @fields_shift[a]) & @fields_mask[a]
                        val = ((val>>8)&0XFF)|((val&0xF)<<8)
                        if val&0x800 == 0x800 then
                            val = -((val^0xFFF)+1)
                        end
                        Metasm::Expression[val+2]
                    end
                    when :disp8 then begin
                        val = (val >> @fields_shift[a]) & @fields_mask[a]
                        if val&0x80 == 0x80 then
                            val = -((val^0xFF)+1)
                        end
                        Metasm::Expression[val+2]
                    end
                    when :i16; Metasm::Expression[(val >> @fields_shift[a]) & @fields_mask[a]]
                    else raise 'fuu'
                    end
                }
           end
        end
		di
	end

	def get_backtrace_binding(di)
		args = di.instruction.args.map { |a|
			case a
			when Reg, MemRef; a.symbolic
			when Metasm::Expression; a
			end
		}

		case di.opcode.name
            when 'add'; { args[0] => Metasm::Expression[args[1], :+, args[2]] }
            when 'lh'; { args[0] => Metasm::Expression[args[1], :<<, 8 ]}
            when 'sub'; { args[0] => Metasm::Expression[args[1], :+, args[2]] }
            when 'xor'; { args[0] => Metasm::Expression[args[1], :^, args[2]] }
            when 'and'; { args[0] => Metasm::Expression[args[1], :&, args[2]] }
            when 'or'; { args[0] => Metasm::Expression[args[1], :|, args[2]] }
            when 'mul'; { args[0] => Metasm::Expression[args[1], :*, args[2]] }
		else {}
		end
	end

	def get_xrefs_x(dasm, di)
		return [] if not di.opcode.props[:setip]
        res = []
		di.instruction.args.each { |a|
			case a
			when Metasm::Expression; res << Metasm::Expression[di.address, :+, a].reduce
			end
		}
        res
	end

	def parse_arg_valid?(op, sym, arg)
		return true	# allow any arg anywhere (for deobf)
		case sym
		when :r_s; arg.kind_of?(Reg)
		when :r_d; arg.kind_of?(Reg)
		when :r_s2; arg.kind_of?(Reg)
		when :i16; arg.kind_of?(Metasm::Expression)
		when :i8; arg.kind_of?(Metasm::Expression)
		end
	end

	def parse_argument(pgm)
		pgm.skip_space
		return if not tok = pgm.nexttok
		if tok.type == :string and tok.raw =~ /^r(\d{1,2})$/
			pgm.readtok
			Reg.new($1.to_i)
		elsif tok.type == :string and tok.raw =~ /^x(.*)$/
			pgm.readtok
            if $1[-1] == 'h'
                Reg.new($1[0..-2].to_i(16)/4+1)
            else
                Reg.new($1.to_i(16)/4+1)
            end
		else
			Metasm::Expression[Metasm::Expression.parse(pgm).reduce]
		end
	end

	def decompile_makestackvars(dasm, funcstart, blocks)
		# no stack
	end

	def decompile_func_finddeps(dcmp, blocks, func)
		# trivial version: assume all blocks should store the final value of all regs (but not the accumulator)
		blocks.inject({}) { |deps, block| deps.update block[0] => (0..15).map { |i| "r#{i}".to_sym } }
	end

	def decompile_blocks(dcmp, myblocks, deps, func, nextaddr=nil)
		# copy/paste from ia32
		scope = func.initializer
		stmts = scope.statements
		until myblocks.empty?
			b, to = myblocks.shift
			if l = dcmp.dasm.get_label_at(b)
				stmts << Metasm::C::Label.new(l)
			end

			ops = []
			binding = {}
			ce = lambda { |*e| dcmp.decompile_cexpr(Metasm::Expression[Metasm::Expression[*e].reduce], scope) }
			ceb = lambda { |*e| ce[Metasm::Expression[*e].bind(binding)] }
			commit = lambda {
				deps[b].map { |k|
					[k, ops.rindex(ops.reverse.find { |r, v| r == k })]
				}.sort_by { |k, i| i.to_i }.each { |k, i|
					next if not i or not binding[k]
					e = k
					final = []
					ops[0..i].reverse_each { |r, v|
						final << r if not v
						e = Metasm::Expression[e].bind(r => v).reduce if not final.include? r
					}
					ops[i][1] = nil
					binding.delete k
					stmts << ce[k, :'=', e] if k != e
				}
			}
			dcmp.dasm.decoded[b].block.list.each_with_index { |di, didx|
				a = di.instruction.args
				if di.opcode.props[:setip] and not di.opcode.props[:stopexec] and to.length > 1
					# conditional jump
					commit[]
					n = dcmp.backtrace_target(get_xrefs_x(dcmp.dasm, di).first, di.address)
					cc = ceb[Metasm::Expression[:acc]]
					stmts << Metasm::C::If.new(Metasm::C::CExpression[cc], Metasm::C::Goto.new(n))
					to.delete dcmp.dasm.normalize(n)
					next
				end

				case di.opcode.name
				when 'ret'
					commit[]
					stmts << Metasm::C::Return.new(nil)
				else
					bd = get_fwdemu_binding(di)
					update = {}
					bd.each { |k, v|
						if k.kind_of? ::Symbol and not deps[b].include? k
							ops << [k, v]
							update[k] = Metasm::Expression[Metasm::Expression[v].bind(binding).reduce]
						else
							stmts << ceb[k, :'=', v]
							stmts.pop if stmts.last.kind_of? Metasm::C::Variable    # [:eflag_s, :=, :unknown].reduce
						end
					}
					binding.update update
				end
			}
			commit[]

			if to.length == 1 and (myblocks.empty? ? nextaddr != to[0] : myblocks.first.first != to[0])
				stmts << Metasm::C::Goto.new(dcmp.dasm.auto_label_at(to[0], 'unknown_goto'))
			end
		end
	end
end

edata = Metasm::EncodedData.new(File.read(ARGV.shift))
dasm = Metasm::Shellcode.decode(edata, SSTIC14_MCU_CPU.new).disassembler
dasm.load_plugin 'deobfuscate'
class << dasm
	Deobfuscate::PatternMacros['%r'] = '(?:r[0-9]{1,2})'
	Deobfuscate::Patterns = {
		'lh (%r), (%i) ; or \1, (%i)' => 'li %1, ((%2<<8)|%3)',
        'call r15.*' =>'ret'
        }
	Deobfuscate.init
end
dasm.disassemble(0)
w = Metasm::Gui::DasmWindow.new("sstic")
w.display(dasm, 0)
Metasm::Gui.main
puts dasm
