require "metasm"
require "pp"

class SSTIC14_CPU < Metasm::CPU
	class Reg
		include Metasm::Renderable

		def initialize(i);  @i = i; end
		def symbolic; "r#@i".to_sym; end
		def render; ["r#@i"]; end
	end

	class Cond
		include Metasm::Renderable
        attr_reader :i
        Flags = ["al", "nv", "z", "nz", "s", "gt", "ge", "7"]

		def initialize(i);  @i = i; end
		def symbolic; Flags[@i].to_sym; end
		def render; [Flags[@i]]; end
	end

	class MemRef
		include Metasm::Renderable
		def initialize(ptr, offset=0); @ptr = ptr; @offset=offset end
		def symbolic;
            Metasm::Indirection[@ptr.kind_of?(Metasm::Expression) ? @ptr : @ptr.symbolic, 1]; 
        end
		def render; 
            res = ['[', @ptr ]
            if @offset.reduce != 0 then
                res << '+'
                res << @offset
            end
            res << ']'
            res
        end
	end

	def initialize
		@size = 32
		@endianness = :little
	end

	def addop(name, bin, *args)
		o = Metasm::Opcode.new(name, bin)
		args.each { |a|
			o.args << a if @valid_args[a]
			o.props[a] = true if @valid_props[a]
			o.fields[a] = [@fields_mask[a], @fields_shift[a]] if @fields_mask[a]
		}
		@opcode_list << o
	end

	def init_opcode_list
		@opcode_list = []

		@valid_props = { :setip => true, :stopexec => true }
		@fields_mask = { :r_d => 0xF, :r_s => 0xF, :r_jmp => 0xF, :i12 => 0xFFFFF, :i16 => 0xFFFF, :flags => 0x7, :i32 => 0xFFFFFFFF}
		@fields_shift = { :r_d => 8, :r_s => 12, :r_jmp => 9, :i16 => 16, :i12 => 12, :flags => 13, :i32 => 8 }
		@valid_args = { :r_d => true, :r_s => true, :r_jmp => true, :i12 => true, :i16 => true, :flags => true, :i32 => true}

        # 4 bytes opcodes
		addop 'lh',  0x00, :r_d, :i12
		addop 'll',  0x01, :r_d, :i12
        # macro instruction
		addop 'li', 0xFF, :r_d, :i32
		addop 'mov.w', 0x02, :r_d, :r_s, :i16
		addop 'mov.b', 0x04, :r_d, :r_s, :i16
		addop 'mov', 0x07, :r_d, :r_s, :i16
        # macro
		addop 'movr', 0xFD, :r_d, :r_s
		addop 'j', 0x08, :r_jmp, :flags, :i16, :setip, :stopexec

        # 2 bytes opcodes
		addop 'xor', 0x0A, :r_d, :r_s
		addop 'or', 0x0B, :r_d, :r_s
		addop 'and', 0x0C, :r_d, :r_s
		addop 'shl', 0x0D, :r_d, :r_s
		addop 'shr', 0x0E, :r_d, :r_s
		addop 'add', 0x12, :r_d, :r_s
		addop 'sub', 0x13, :r_d, :r_s

		addop 'inc', 0x16, :r_d
		addop 'dec', 0x17, :r_d

		addop 'exit', 0x1c, :stopexec
		addop 'read_write', 0x1d

		addop 'parity', 0x1e, :r_d, :r_s
	end
	
	def build_bin_lookaside
		opcode_list.each { |op|
			mask = 0
			op.fields.each { |fld, (msk, shft)| mask |= msk << shft }
			op.bin_mask = ~mask
		}
	end

	def decode_findopcode(edata)
		di = Metasm::DecodedInstruction.new(self)
		val = edata.decode_imm(:u8, @endianness)
		edata.ptr -= 1
		if op = opcode_list.find { |o| val & o.bin_mask == o.bin }
			di.opcode = op
            if val <= 8
                di.bin_length = 4 
            else
                di.bin_length = 2
            end
        else
            puts val
            puts (edata.ptr+0x40).to_s(16)
            raise "fuu"
		end
        di
	end

	def decode_instr_op(edata, di)
        if di.bin_length == 4 then
            val = edata.decode_imm(:u32, @endianness)
        else
            val = edata.decode_imm(:u16, @endianness)
        end
		di.instruction.opname = di.opcode.name

        case di.opcode.name 
            when "mov.b" then begin
                di.instruction.args << Reg.new((val >> @fields_shift[:r_d]) & @fields_mask[:r_d])
                di.instruction.args << MemRef.new(Reg.new((val >> @fields_shift[:r_s]) & @fields_mask[:r_s]), Metasm::Expression[(val >> @fields_shift[:i16]) & @fields_mask[:i16]])
            end
            when "mov.w" then begin
                di.instruction.args << Reg.new((val >> @fields_shift[:r_d]) & @fields_mask[:r_d])
                di.instruction.args << MemRef.new(Reg.new((val >> @fields_shift[:r_s]) & @fields_mask[:r_s]), Metasm::Expression[(val >> @fields_shift[:i16]) & @fields_mask[:i16]])
            end
            when "mov" then begin
                di.instruction.args << MemRef.new(Reg.new((val >> @fields_shift[:r_s]) & @fields_mask[:r_s]), Metasm::Expression[(val >> @fields_shift[:i16]) & @fields_mask[:i16]])
                di.instruction.args << Reg.new((val >> @fields_shift[:r_d]) & @fields_mask[:r_d])
            end
            else begin
                di.opcode.args.each { |a|
                    di.instruction.args << case a
                    when :r_s; Reg.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :r_d; Reg.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :r_jmp; Reg.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :flags; Cond.new((val >> @fields_shift[a]) & @fields_mask[a])
                    when :i12; Metasm::Expression[(val >> @fields_shift[a]) & @fields_mask[a]]
                    when :i16; Metasm::Expression[(val >> @fields_shift[a]) & @fields_mask[a]]
                    else raise 'fuu'
                    end
                }
           end
        end
		di
	end

	def get_backtrace_binding(di)
		args = di.instruction.args.map { |a|
			case a
			when Reg, MemRef; a.symbolic
			when Metasm::Expression; a
			end
		}

		case di.opcode.name
		when 'mov'; { args[0] => Metasm::Expression[args[1]] }
		when 'and'; { args[0] => Metasm::Expression[args[0], :&, args[1]] }
		when 'or';  { args[0] => Metasm::Expression[args[0], :|, args[1]] }
		when 'xor';  { args[0] => Metasm::Expression[args[0], :^, args[1]] }
		when 'not'; { args[0] => Metasm::Expression[args[0], :^, Metasm::Expression[0xffff]] }
		when 'shl'; { args[0] => Metasm::Expression[args[0], :<<, 1] }
		when 'shr'; { args[0] => Metasm::Expression[args[0], :>>, 1] }
		else {}
		end
	end

	def get_xrefs_x(dasm, di)
		return [] if not di.opcode.props[:setip]
        res = []
		args = di.instruction.args.each { |a|
			case a
			when Metasm::Expression; res << a
			when Cond then begin
                if a.i != 0
                    res  << [[di.address+4]]
                end
            end
			end
		}
        return res
	end

	def parse_arg_valid?(op, sym, arg)
		return true	# allow any arg anywhere (for deobf)
		case sym
		when :r_s; arg.kind_of?(Reg)
		when :r_d; arg.kind_of?(MemRef)
		when :r_jmp; arg.kind_of?(MemRef)
		when :i16; arg.kind_of?(Metasm::Expression)
		when :i12; arg.kind_of?(Metasm::Expression)
		end
	end

	def parse_argument(pgm)
		pgm.skip_space
		return if not tok = pgm.nexttok
		if tok.type == :string and tok.raw =~ /^r(\d{1,2})$/
			pgm.readtok
			Reg.new($1.to_i)
		elsif tok.type == :string and tok.raw =~ /^x(.*)$/
			pgm.readtok
            if $1[-1] == 'h'
                Reg.new($1[0..-2].to_i(16)/4+1)
            else
                Reg.new($1.to_i(16)/4+1)
            end
		else
			Metasm::Expression[Metasm::Expression.parse(pgm).reduce]
		end
	end

	def decompile_makestackvars(dasm, funcstart, blocks)
		# no stack
	end

	def decompile_func_finddeps(dcmp, blocks, func)
		# trivial version: assume all blocks should store the final value of all regs (but not the accumulator)
		blocks.inject({}) { |deps, block| deps.update block[0] => (0..7).map { |i| "r#{i}".to_sym } }
	end

	def decompile_blocks(dcmp, myblocks, deps, func, nextaddr=nil)
		# copy/paste from ia32
		scope = func.initializer
		stmts = scope.statements
		until myblocks.empty?
			b, to = myblocks.shift
			if l = dcmp.dasm.get_label_at(b)
				stmts << Metasm::C::Label.new(l)
			end

			ops = []
			binding = {}
			ce = lambda { |*e| dcmp.decompile_cexpr(Metasm::Expression[Metasm::Expression[*e].reduce], scope) }
			ceb = lambda { |*e| ce[Metasm::Expression[*e].bind(binding)] }
			commit = lambda {
				deps[b].map { |k|
					[k, ops.rindex(ops.reverse.find { |r, v| r == k })]
				}.sort_by { |k, i| i.to_i }.each { |k, i|
					next if not i or not binding[k]
					e = k
					final = []
					ops[0..i].reverse_each { |r, v|
						final << r if not v
						e = Metasm::Expression[e].bind(r => v).reduce if not final.include? r
					}
					ops[i][1] = nil
					binding.delete k
					stmts << ce[k, :'=', e] if k != e
				}
			}
			dcmp.dasm.decoded[b].block.list.each_with_index { |di, didx|
				a = di.instruction.args
				if di.opcode.props[:setip] and not di.opcode.props[:stopexec] and to.length > 1
					# conditional jump
					commit[]
					n = dcmp.backtrace_target(get_xrefs_x(dcmp.dasm, di).first, di.address)
					cc = ceb[Metasm::Expression[:acc]]
					stmts << Metasm::C::If.new(Metasm::C::CExpression[cc], Metasm::C::Goto.new(n))
					to.delete dcmp.dasm.normalize(n)
					next
				end

				case di.opcode.name
				when 'ret'
					commit[]
					stmts << Metasm::C::Return.new(nil)
				else
					bd = get_fwdemu_binding(di)
					update = {}
					bd.each { |k, v|
						if k.kind_of? ::Symbol and not deps[b].include? k
							ops << [k, v]
							update[k] = Metasm::Expression[Metasm::Expression[v].bind(binding).reduce]
						else
							stmts << ceb[k, :'=', v]
							stmts.pop if stmts.last.kind_of? Metasm::C::Variable    # [:eflag_s, :=, :unknown].reduce
						end
					}
					binding.update update
				end
			}
			commit[]

			if to.length == 1 and (myblocks.empty? ? nextaddr != to[0] : myblocks.first.first != to[0])
				stmts << Metasm::C::Goto.new(dcmp.dasm.auto_label_at(to[0], 'unknown_goto'))
			end
		end
	end
end

edata = Metasm::EncodedData.new(File.read(ARGV.shift))
dasm = Metasm::Shellcode.decode(edata, SSTIC14_CPU.new).disassembler
dasm.load_plugin 'deobfuscate'
class << dasm
	Deobfuscate::PatternMacros['%r'] = '(?:r[0-9]{1,2})'
	Deobfuscate::Patterns = {
		'mov\.w (%r), \[r0\]' => 'movr %1, r1',
		'lh (%r), (%i) ; ll \1, (%i)' => 'li %1, ((%2<<16)|%3)',
		'mov\.w (%r), \[r0\+(%i)\]' => 'movr %1, x%2'
        }
	Deobfuscate.init
end
dasm.disassemble(0x40)
w = Metasm::Gui::DasmWindow.new("sstic")
w.display(dasm, 0x40)
Metasm::Gui.main
